/*
 * Slightly smarter Micro-C ROBOT
 *
 * Compile command: ccf killer -fop
 */
#include <cflea.h>
#include "mcrobots.h"

#define	BOUNDARY	20			/* Reverse this many points from edge */
#define	DAM_MOVE	100			/* Avoid damage for this many ticks */
#define	IDLE_MOVE	777			/* Change direction if idle for this long */

/* Travel direction vectors (subtract 1 for scan/shoot vectors)
 *             N
 *       NNW   1  NNE
 *     NW  16     2  NE
 *  WNW  15         3  ENE
 *     14             4
 * W 13        0        5 E
 *     12             6
 *  WSW  11         7  ESE
 *     SW  10     8  SE
 *       SSW   9  SSE
 *             S
 */
#define	STOP	0
#define	N		1		/* North */
#define	NNE		2		/* North NorthEast */
#define	NE		3		/* NorthEast */
#define	ENE		4		/* East NorthEast */
#define	E		5		/* East */
#define	ESE		6		/* Easy SouthEast */
#define	SE		7		/* SouthEast */
#define	SSE		8		/* South SouthEast */
#define	S		9		/* South */
#define	SSW		10		/* South SouthWest */
#define	SW		11		/* South West */
#define	WSW		12		/* West SouthWest */
#define	W		13		/* West */
#define	WNW		14		/* West NorthWest */
#define	NW		15		/* NorthWest */
#define	NNW		16		/* North NorthWest */

/* Travel reversal tables for X and Y axis:
 *	0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16
 * STOP	N  NNE NE  ENE	E  ESE SE  SSE	S SSW  SW WSW   W WNW  NW  NNW */
static unsigned reverse_x[] = {		/* Table to reverse X direction */
   STOP,N, NNW,NW, WNW, W, WSW,SW, SSW, S, SSE,SE, ESE, E, ENE,NE, NNE };
static unsigned reverse_y[] = {		/* Table to reverse Y direction */
   STOP,S, SSE,SE, ESE, E, ENE,NE, NNE, N, NNW,NW,WNW, W, WSW, SW, SSW };

/*
 * Accepts a 0-based direction (0-(MAX_DIR-1)), and adjusts it so that
 * it will not take us into a nearby wall. Returns a 1-based direction
 * suitable for set_direction()
 */
int avoid_wall(int d)
{
	unsigned x, y;
	char dflag;

	x = get_x();
	y = get_y();
	do {
		dflag = 0;
		d &= (MAX_DIR-1);
		switch(d) {	/* Check too far north? */
		case WNW-1	:
		case NW-1	:
		case NNW-1	:
		case N-1	:
		case NNE-1	:
		case NE-1	:
		case ENE-1	:
			if(y < BOUNDARY) dflag = -1; }
		switch(d) {	/* Check too far west ? */
		case SSW-1	:
		case SW-1	:
		case WSW-1	:
		case W-1	:
		case WNW-1	:
		case NW-1	:
		case NNW-1	:
			if(x < BOUNDARY) dflag = -1; }
		switch(d) {	/* Check too far east? */
		case NNE-1	:
		case NE-1	:
		case ENE-1	:
		case E-1	:
		case ESE-1	:
		case SE-1	:
		case SSE-1	:
			if(x > (X_ARENA-BOUNDARY)) dflag = -1; }
		switch(d) {	/* Check to far south? */
		case ESE-1	:
		case SE-1	:
		case SSE-1	:
		case S-1	:
		case SSW-1	:
		case SW-1	:
		case WSW-1	:
			if(y > (Y_ARENA-BOUNDARY)) dflag = -1; }
		++d; }
	while(dflag);
	return d;
}

/*
 * Main "killer" robot program
 */
main()
{
    unsigned wd, sd, i, x, y, b, daflag, imflag;
	char hunt, xfix, yfix;

	printf("KILLER is ready for action!\n");

	b = (get_id() * 200) + 3000;	/* Unique tone for this robot */
	i = hunt = daflag = 0;
	xfix = yfix = -1;
	imflag = IDLE_MOVE;

	set_direction(wd = avoid_wall((sd = rand(MAX_DIR)) + 1));

	for(;;) {
		/*
		 * Movement logic, if we are within BOUNDARY of any edge,
		 * reverse direction in that axis, make sure we don't reverse
		 * again until we are clear.
		 */
		x = get_x();
		y = get_y();
		if((x < BOUNDARY) || (x > (X_ARENA - BOUNDARY))) {
			if(xfix) {
				set_direction(wd = reverse_x[wd]);
				xfix = 0; } }
		else
			xfix = -1;

		if((y < BOUNDARY) || (y > (Y_ARENA - BOUNDARY))) {
			if(yfix) {
				set_direction(wd = reverse_y[wd]);
				yfix = 0; } }
		else
			yfix = -1;

		/*
		 * Damage avoidance... If we detect that we are
		 * incurring damage, change our direction at right angles.
		 */
        if(get_damage() != i) {	/* We are getting hit! */
			daflag = DAM_MOVE;
			x = get_x();
			y = get_y();
			set_direction(wd = avoid_wall(wd+4));
            i = get_damage();
			beep(b);		/* Scream ... */
			beep(b+100);	/* when we... */
			beep(b); }		/* get hurt!! */

		/*
		 * Shooting logic... If we are scanning a target in the
		 * current scan/shoot direction, then head for it and blast
		 * it. Otherwise, if we have just lost the target, reset our
		 * scan angle to try and re-scan that area, otherwise, just
		 * rotate the scan angle around till we see another target.
		 */
        if(scan(sd)) {			/* Found someone! */
			imflag = IDLE_MOVE;
            shoot(sd);
			if(!daflag)
				set_direction(wd = avoid_wall(sd));
			hunt = -1; }
        else if(hunt) {
			sd = (sd - 4) & (MAX_DIR-1);
			hunt = 0; }
		else {
			if(!--imflag) {
				printf("Oh where.. Oh where.. have my enemies gone?\n");
				set_direction(wd = avoid_wall(wd+6));
				imflag = IDLE_MOVE; }
			sd = (sd + 1) & (MAX_DIR-1); }

		if(daflag)
			--daflag; }
}
