/*
 * Micro-C ROBOTS using a multi-tasking C-FLEA virtual machine
 *
 * This program must be compiled with the Micro-C/PC compiler, which is
 * available from http://www.dunfield.com, and is also inclued on the DDS
 * Demo/Support diskette.
 *
 * Compile command: cc MCROBOTS -fop
 *
 * Copyright 1996-2005 Dave Dunfield
 * All rights reserved.
 */
#include <stdio.h>
#include <window.h>

/*
 * MCROBOTS general fixed parameters
 */
#define X_ARENA 	128			/* Size of actual area */
#define Y_ARENA 	64
#define X_SIZE		64			/* Size of viewable area */
#define Y_SIZE		16
#define	X_START		20			/* Size of starting blocks */
#define	Y_START		10
#define MAX_SHOTS	50			/* Max shots */
#define	MAX_DIR		16			/* Maximum direction */
#define	MAX_SPEED	10			/* Maximum speed reached by robot */
#define	MAX_TASKS	8			/* Maximum number of robots */

/*
 * C-FLEA internal definitions
 */
#define	BYTE_OP		0x08		/* Bit in opcode indicating byte operand */
#define	EQ			0x01		/* Bit in cflags indicating == */
#define	ULT			0x02		/* Bit in cflags indicating unsigned < */
#define	UGT			0x04		/* Bit in cflags indicating unsigned > */
#define	SLT			0x08		/* Bit in cflags indicating signed < */
#define	SGT			0x10		/* Bit in cflags indicating signed > */

/*
 * Sound definitions
 */
#define	SND_QUEUE	64			/* Size of sound queue (power of 2) */
#define	SND_DIE1	150			/* Tone when robots die */
#define	SND_DIE2	125			/* "" */
#define	SND_DIE3	100			/* "" */
#define	SND_HIT		300			/* Tone when robot gets hit */
#define	SND_WALL1	250			/* Tone when robot hits wall */
#define	SND_WALL2	275			/* "" */
#define	SND_CRASH1	225			/* Tone when robots collide */
#define	SND_CRASH2	250			/* "" */
#define	SND_SHOT	2500		/* Tone when robot shoots */

/*
 * Adjustable system operational parameters
 */
static unsigned
	Slottime = 250,				/* Number of instructions per slot */
	Maxdamage = 100,			/* Maximum damage before failure */
	Scan_dist = 25,				/* Scan distance */
	Scan_area = 5,				/* Scan area */
	Shot_dist = 20,				/* Shot distance */
	Shot_area = 5,				/* Shot area */
	Shot_delay = 20,			/* Shot delay */
static unsigned char
	Win = -1,					/* Detect winning */
	Rstart = -1,				/* Random start */
	Sound = -1,					/* Sound is enabled */
	Inter = -1;					/* Interactive setup inhibit */

/*
 * Misc. global variables
 */
static unsigned
	Task,						/* Currently active task */
	Tick = 0,					/* System tick count */
	arena_seg,					/* Segment for arena */
	Address,					/* C-FLEA Address holding register */
	Sound_queue[SND_QUEUE],		/* Sound queue data */
	Sound_head = 0,				/* Head of sound queue */
	Sound_tail = 0;				/* Tail of sound queue */
static unsigned char
	opcode,						/* C-FLEA opcode being decoded */
	*ptr;						/* General purpose parsing pointer */
struct WINDOW
	*Awin,						/* Arena window handle */
	*Swin,						/* Status window handle */
	*Mwin;						/* Message window handle */
FILE
	*mfp;						/* Message file pointer */

/*
 * Form for interactive operational parameter setup
 */
char *data_form[] = {
	32<<8|9,
	"\x01\x00\x83Instructions/time slot:",
	"\x01\x01\x83Maximum damage        :",
	"\x01\x02\x82Scanning distance     :",
	"\x01\x03\x82Scanning area         :",
	"\x01\x04\x82Shooting distance     :",
	"\x01\x05\x82Shooting area         :",
	"\x01\x06\x83Delay between shots   :",
	0 };

/*
 * Active shot data
 */
unsigned
	shots_x[MAX_SHOTS],			/* Shot X position */
	shots_y[MAX_SHOTS],			/* Shot Y position */
	shots_d[MAX_SHOTS],			/* Shot direction */
	shots_v[MAX_SHOTS],			/* Shot velocity */
	shots_o[MAX_SHOTS];			/* Shot owner */

/* Scan/Shoot direction vectors (add 1 for travel vectors)
 *           00
 *        15     01
 *     14          02
 *   13              03
 * 12                  04
 *   11              05
 *     10          06
 *        09    07
 *           08
 *				0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15 */
int delta_x[]={ 0,  1,  1,  2,  2,  2,  1,  1,  0, -1, -1, -2, -2, -2, -1, -1 };
int delta_y[]={-2, -2, -1, -1,  0,  1,  1,  2,  2,  2,  1,  1,  0, -1, -1, -2 };

/*
 * Video/Window attributes
 */
unsigned
	cattrs[] = {		/* Color video attributes */
		WCOPEN|WCCLOSE|0x6E,			/* 0 = Arena */
		WCOPEN|WCCLOSE|0x47,			/* 1 = Status window */
		WCOPEN|WCCLOSE|WSCROLL|0x17,	/* 2 = Message window */
		WSAVE|WCOPEN|WBOX1|REVERSE },	/* 3 = Setup form */
	mattrs[] = {		/* Monochrome video attributes */
		WCOPEN|WCCLOSE|NORMAL,			/* 0 = Arena */
		WCOPEN|WCCLOSE|REVERSE,			/* 1 = Status window */
		WCOPEN|WCCLOSE|WSCROLL|REVERSE,	/* 2 = Message window */
		WSAVE|WCOPEN|WBOX1|REVERSE },	/* 3 = Setup form */
	*attrs = &cattrs;	/* Default to color attributes */

/*
 * Help text
 */
static char help[] = { "\n\
Micro-C Robots!\n\n\
Use: mcrobots robot_file(s)... [options]\n\n\
opts:	-I	= no Interactive setup screen\n\
	-R	= do not Randomize start positions\n\
	-S	= no Sound\n\
	-W	= do not detect Winning\n\
	A=n	= set scAn area\n\
	C=n	= set sCan distance\n\
	F=file	= write message log to File\n\
	H=n	= set sHot distance\n\
	I=n	= set Instructions/time slot\n\
	M=n	= set Maximum damage\n\
	O=n	= set shOt area\n\
	T=n	= set shoT delay time\n\n\
Copyright 1999-2005 Dave dunfield\n\
All rights reserved.\n" };

/* Random seed is accessed to randomize the generator */
extern unsigned RAND_SEED;

/*
 * Robot control structure
 *
 * This contains all data required for the C-FLEA virtual machine
 * and the robot control functions.
 */
struct CF_data {
	/* C-FLEA virtual machine variables */
	unsigned _ACC;				/* C-FLEA accumulator */
	unsigned _INDEX;			/* C-FLEA index register */
	unsigned _SP;				/* C-FLEA stack pointer */
	unsigned _PC;				/* C-FLEA program counter */
	unsigned _CFSEG;			/* C-FLEA code segment */
	unsigned _Alt;				/* C-FLEA alternate register */
	unsigned char _cflags;		/* C-FLEA condition flags register */
	/* Robot control variables */
	unsigned _X, _Y;			/* X and Y positions */
	unsigned _Direct;			/* Current direction */
	unsigned _Ldirect;			/* Last direction */
	unsigned _Speed;			/* Current speed */
	unsigned _Damage;			/* Accumulated damage */
	unsigned _Shot;				/* last shot count */
	/* Misc housekeeping variables */
	char *_Name;				/* Robot name */
	unsigned _Pptr;				/* Current print position */
	unsigned char _Pbuf[81];	/* Print output buffer */
	} *CF_ptr;

/*
 * These defines provide the appearance of simple variable access
 * to elements of the currently active robot control structure.
 */
#define	ACC		(CF_ptr->_ACC)
#define	INDEX	(CF_ptr->_INDEX)
#define	SP		(CF_ptr->_SP)
#define	PC		(CF_ptr->_PC)
#define	CFSEG	(CF_ptr->_CFSEG)
#define	Alt		(CF_ptr->_Alt)
#define	cflags	(CF_ptr->_cflags)

#define	X		(CF_ptr->_X)
#define	Y		(CF_ptr->_Y)
#define	Direct	(CF_ptr->_Direct)
#define	Ldirect	(CF_ptr->_Ldirect)
#define	Speed	(CF_ptr->_Speed)
#define	Damage	(CF_ptr->_Damage)
#define Shot	(CF_ptr->_Shot)

#define	Name	(CF_ptr->_Name)
#define	Pptr	(CF_ptr->_Pptr)
#define	Pbuf	(CF_ptr->_Pbuf)

struct CF_data
	CF_block[MAX_TASKS];		/* Array of active task blocks */
unsigned
	CF_tasks = 0;				/* Number of tasks loaded */

/*
 * Functions to read and write C-FLEA memory
 */
#define	read_mem(address)			peek(CFSEG, address)
#define	write_mem(address, data)	poke(CFSEG, address, data)

/*
 * Formatted print of an informational message
 */
register message(unsigned args)
{
	char buffer[81];

	_format_(nargs() * 2 + &args, buffer);
	wprintf("\n[%u] %s", Task+1, buffer);
	if(mfp)
		fprintf(mfp, "[%u] %s\n", Task+1, buffer);
}

/*
 * Formatted print of error message & exit
 */
register error(unsigned args)
{
	char buffer[81];

	_format_(nargs() * 2 + &args, buffer);

	*Mwin = *Swin = *Awin = NORMAL;
	wclose();
	wclose();
	wclose();
	printf("[%u] %s\n", Task+1, buffer);
	exit(-1);
}

/*
 * Test for HEX digit & return value
 */
static unsigned get_hex(int c)
{
	if((c >= '0') && (c <= '9'))
		return c-'0';
	if((c >= 'A') && (c <= 'F'))
		return c-'A'+10;
	if((c >= 'a') && (c <= 'f'))
		return c-'a'+10;
	error("Bad hex digit");
}

/*
 * Load a byte from the hex file
 */
static unsigned get_byte(void)
{
	return (get_hex(*ptr++) << 4) + get_hex(*ptr++);
}

/*
 * Read file into memory image (Intel or Motorola format)
 */
unsigned read_file(char *name)
{
	unsigned a, b, chksum, loadptr, lcount, count, maxsize, base;
	char buffer[100], dot;
	FILE *fp;

	dot = -1;
	ptr = buffer;
	while(*name) {
		*ptr++ = *name = toupper(*name);
		if(*name == '.')
			*name = dot = 0;
		++name; }
	*ptr = 0;
	if(dot)
		strcpy(ptr, ".HEX");

	if(!(fp = fopen(buffer, "r")))
		error("Unable to read file: %s", buffer);

	lcount = maxsize = 0;
	base = -1;
	for(;;) {
		if(!fgets(ptr = buffer, sizeof(buffer)-1, fp)) {
			message("No end of file record.");
			goto quit; }
		again: switch(*ptr++) {
			case 'S' :	/* Motorola HEX format */
				if(*ptr == '9') goto quit;
				if(*ptr++ != '1') continue;
				lcount += count = (chksum = get_byte()) - 3;
				chksum += a = get_byte();
				chksum += b = get_byte();
				loadptr = (a << 8) + b;
				if(base == -1)
					base = loadptr;
				while(count--) {
					chksum += a = get_byte();
					write_mem(loadptr++, a); }
				if((255 & ~chksum) != get_byte())
					goto quitbad;
				break;
			case ':' :		/* Intel HEX format */
				if(!(count = get_byte())) goto quit;
				lcount += count;
				chksum = (a = get_byte()) + count;
				chksum += b = get_byte();
				loadptr = (a << 8) + b;
				if(base == -1)
					base = loadptr;
				chksum += get_byte();
				while(count--) {
					chksum += a = get_byte();
					write_mem(loadptr++, a); }
				if((255 & -chksum) != get_byte())
					goto quitbad;
				break;
			case ' ' :	/* Space */
			case '\t' :	/* Tab */
				goto again;
			case 0 :		/* Null line */
				continue;
			default:
				error("Invalid record format"); }
		if(loadptr > maxsize)
			maxsize = loadptr; }

quitbad:
	error("Bad record checksum");
quit:
	fclose(fp);

	PC = base;

	return maxsize;
}

/* --- Multitasking C-FLEA Virtual Machine Implementation --- */

/*
 * Compute the address of the current instructions operand
 */
static void cf_get_address(void)
{
	switch(opcode & 7) {
		case 0 :		/* Immediate reference */
			Address = PC++;
			return;
		case 1 :		/* Direct address */
			Address = read_mem(PC++);
			Address |= read_mem(PC++) << 8;
			return;
		case 2 :		/* Indirect - no offset */
			Address = INDEX;
			return;
		case 3 :		/* Indirect with offset */
			Address = INDEX + read_mem(PC++);
			return;
		case 4 :		/* Indirect from SP with offset */
			Address = SP + read_mem(PC++);
			return;
		case 5 :		/* On top of stack, remove */
			Address = SP;
			SP += 2;
			return;
		case 6 :		/* Indirect through top of stack, remove */
			Address = read_mem(SP++);
			Address |= read_mem(SP++) << 8;
			return;
		case 7 :
			Address = read_mem(SP);
			Address |= read_mem(SP+1) << 8;
			return; }
}

/*
 * Main C-FLEA virtual machine execution function
 */
unsigned cf_run(void)
{
	unsigned operand, t1, t2, slot;

	slot = Slottime;
	do {
		opcode = read_mem(PC++);

		if(opcode < 0x98) {				/* Catagory 1 instruction */
			cf_get_address();
			operand = read_mem(Address);
			if(!(opcode & BYTE_OP)) {
				if(!(opcode & 7))
					++PC;
				operand |= read_mem(Address+1) << 8; }
			switch(opcode & 0xF0) {
			case 0x00 :	ACC = operand;					break;	/* LD */
			case 0x10 : ACC += operand;					break;	/* ADD */
			case 0x20 : ACC -= operand;					break;	/* SUB */
			case 0x30 : ACC *= operand;					break;	/* MUL */
			case 0x40 :
				Alt = ACC % operand; ACC /= operand;	break;	/* DIV */
			case 0x50 : ACC &= operand;					break;	/* AND */
			case 0x60 :	ACC |= operand;					break;	/* OR */
			case 0x70 :	ACC ^= operand;					break;	/* XOR */
			case 0x80 :											/* CMP */
				if(ACC == operand) {
					cflags = EQ;
					ACC = 1; }
				else {
					cflags = (ACC < operand) ? ULT : UGT;
					cflags |= ((int)ACC < (int)operand) ? SLT : SGT;
					ACC = 0; }
				break;
			case 0x90 :	INDEX = operand; }						/* LDI */
		} else if(opcode < 0xC8) {		/* Catagory 2 instruction */
			cf_get_address();
			switch(opcode & 0xF8) {
			case 0x98 :	INDEX = Address;				break;	/* LEAI */
			case 0xA0 :	write_mem(Address+1, ACC >> 8);			/* ST */
			case 0xA8 :	write_mem(Address, ACC);		break;	/* STB */
			case 0xB0 :	write_mem(Address, INDEX);				/* STI */
						write_mem(Address+1, INDEX >> 8); break;
			case 0xB8 : ACC >>= read_mem(Address);		break;	/* SHR */
			case 0xC0 :	ACC <<= read_mem(Address); }			/* SHL */
		} else switch(opcode) {			/* Catagory 3 instruction */
		case 0xC8 :	ACC = (cflags & (SLT))	  ? 1 : 0;	break;	/* LT */
		case 0xC9 : ACC = (cflags & (SLT|EQ)) ? 1 : 0;	break;	/* LE */
		case 0xCA : ACC = (cflags & (SGT))	  ? 1 : 0;	break;	/* GT */
		case 0xCB : ACC = (cflags & (SGT|EQ)) ? 1 : 0;	break;	/* GE */
		case 0xCC : ACC = (cflags & (ULT))	  ? 1 : 0;	break;	/* ULT */
		case 0xCD : ACC = (cflags & (ULT|EQ)) ? 1 : 0;	break;	/* ULE */
		case 0xCE : ACC = (cflags & (UGT))	  ? 1 : 0;	break;	/* UGT */
		case 0xCF : ACC = (cflags & (UGT|EQ)) ? 1 : 0;	break;	/* UGE */
		case 0xD0 :												/* JMP */
		doljmp:		Address = read_mem(PC++);
					PC = (read_mem(PC++) << 8) | Address;break;
		case 0xD1 : if(!ACC) goto doljmp; PC+=2;		break;	/* JZ */
		case 0xD2 :	if(ACC)  goto doljmp; PC+=2;		break;	/* JNZ */
		case 0xD3 :
		dosjmp:		Address = read_mem(PC++);					/* SJMP */
					if(Address & 0x80) Address |= 0xFF00;
					PC += Address;						break;
		case 0xD4 :	if(!ACC) goto dosjmp; ++PC;			break;	/* SJZ */
		case 0xD5 :	if(ACC)  goto dosjmp; ++PC;			break;	/* SJNZ */
		case 0xD6 :	PC = ACC;							break;	/* IJMP */
		case 0xD7 :												/* SWITCH */
			Address = INDEX;
			for(;;) {
				t1 = read_mem(Address++);
				t1 |= read_mem(Address++) << 8;
				t2 = read_mem(Address++);
				t2 |= read_mem(Address++) << 8;
				if(!t1) {
					PC = t2;
					break; }
				if(t2 == ACC) {
					PC = t1;
					break; } }
			break;
		case 0xD8 :												/* CALL */
			operand = PC + 2;
			write_mem(--SP, operand >> 8);
			write_mem(--SP, operand);
			goto doljmp;
		case 0xD9 :												/* RET */
			PC = read_mem(SP++);
			PC |= read_mem(SP++) << 8;					break;
		case 0xDA : SP -= read_mem(PC++);				break;	/* ALLOC */
		case 0xDB :	SP += read_mem(PC++);				break;	/* FREE */
		case 0xDC :												/* PUSHA */
			write_mem(--SP, ACC >> 8);
			write_mem(--SP, ACC);						break;
		case 0xDD :												/* PUSHI */
			write_mem(--SP, INDEX >> 8);
			write_mem(--SP, INDEX);						break;
		case 0xDE : SP = ACC;							break;	/* TAS */
		case 0xDF : ACC = SP;							break;	/* TSA */
		case 0xE0 : ACC = 0;							break;	/* CLR */
		case 0xE1 : ACC = ~ACC;							break;	/* COM */
		case 0xE2 : ACC = -ACC;							break;	/* NEG */
		case 0xE3 : ACC = !ACC;							break;	/* NOT */
		case 0xE4 : ++ACC;								break;	/* INC */
		case 0xE5 : --ACC;								break;	/* DEC */
		case 0xE6 : INDEX = ACC;						break;	/* TAI */
		case 0xE7 : ACC = INDEX;						break;	/* TIA */
		case 0xE8 : INDEX += ACC;						break;	/* ADAI */
		case 0xE9 : ACC = Alt;							break;	/* ALT */
		case 0xEA :												/* OUT */
			if(cf_output(read_mem(PC++)))
				goto suicide;
			break;
		case 0xEB :												/* IN */
			if(cf_input(read_mem(PC++)))
				goto suicide;
			break;
		default:
			message("Unknown opcode %02x at %04x", opcode, PC-2);
		suicide:
			fall_on_sword();
			return; } }
	while(--slot);
}

/*
 * This function is called in response to an OUT instruction
 */
int cf_output(unsigned port)
{
	switch(port) {
		case 0 :	/* Console output */
			if(ACC == '\n') {
				Pbuf[Pptr] = 0;
				message(Pbuf);
				Pptr = 0; }
			else if(Pptr < 80)
				Pbuf[Pptr++] = ACC;
			return 0;
		case 1 :	/* Set direction */
			if((Direct = ACC) > MAX_DIR) {
				message("Invalid: set_direction(%u)", Ldirect);
				return -1; }
			return 0;
		case 2:		/* do shot */
			if(ACC >= MAX_DIR) {
				message("Invalid: shoot(%u)", ACC);
				return -1; }
			add_shot(ACC);
			return 0;
		case 3 :	/* Warp */
			w_gotoxy(X / (X_ARENA/X_SIZE), Y / (Y_ARENA/Y_SIZE), Awin);
			w_putc(' ', Awin);
			if( ((X = ACC>>8) >= X_ARENA) || ((Y = ACC&255) >= Y_ARENA)) {
				message("Invalid: warp(%u,%u)", X, Y);
				X = Y = 0;
				return -1; }
			Shot = Shot_delay * 5;
			return 0;
		case 4 :	/* Sound */
			add_sound(ACC);
			return 0; }

	message("Unknown output port %02x at %04x", port, PC-2);
	return -1;
}

/*
 * This function is called in response to an IN instruction
 */
int cf_input(unsigned port)
{
	switch(port) {
		case 0 : ACC = Task+1;	return 0;		/* Get our id */
		case 1 : ACC = Direct;	return 0;		/* Read direction */
		case 2 : ACC = Speed;	return 0;		/* Read speed */
		case 3 : ACC = X;		return 0;		/* Read X position */
		case 4 : ACC = Y;		return 0;		/* Read Y position */
		case 5 : ACC = Damage;	return 0;		/* Read Damage */
		case 6 : ACC = Shot;	return 0;		/* Read shot delay */
		case 7 :								/* Perform scan */
			if(ACC >= MAX_DIR) {
				message("Invalid: scan(%u)", ACC);
				return -1; }
			ACC = scan(ACC);
			return 0; }

	message("Unknown input port %02x at %04x", port, PC-2);
	return -1;
}

/* --- Robot control functions --- */

/*
 * Robot has done something foolish/disallowed - kill him!
 */
fall_on_sword()
{
	message("Tripped and fell on his sword! (PC=%04x)", PC);
	PC = 0;
}

/*
 * Cause damage to the robot
 */
damage(unsigned robot, unsigned amount)
{
	unsigned Stask;
	struct CF_data *Sptr;

	Stask = Task;
	Sptr = CF_ptr;
	CF_ptr = CF_block[Task = robot];

	if(PC) {
		Damage += amount;
		message("takes %u damage, Total=%u", amount, Damage);
		Speed = Direct = Ldirect = 0;
		if(Damage >= Maxdamage) {
			add_sound(SND_DIE1);
			add_sound(SND_DIE2);
			add_sound(SND_DIE3);
			message("Has died!");
			PC = 0; } }
	Task = Stask;
	CF_ptr = Sptr;
}

/*
 * Move the robot using his current direction and speed
 */
move_robot()
{
	int dx, dy;
	unsigned char c;

	if(Direct != Ldirect) {		/* Robot has changed direction */
		if(Speed)
			--Speed;
		else
			Ldirect = Direct; }
	else if(Speed < MAX_SPEED)
		++Speed;

	if(Tick > Speed)			/* Handle acceleration */
		return;

	if(Ldirect) {
		dx = delta_x[Ldirect-1];
		dy = delta_y[Ldirect-1];

		w_gotoxy(X / (X_ARENA/X_SIZE), Y / (Y_ARENA/Y_SIZE), Awin);
		w_putc(' ', Awin);
		poke(arena_seg, (X*Y_ARENA)|Y, 0);

		if((X += dx) >= X_ARENA) {
			message("Hit the wall");
			X -= dx;
			add_sound(SND_WALL1);
			add_sound(SND_WALL2);
			damage(Task, Speed); }
		if((Y += dy) >= Y_ARENA) {
			message("Hit the wall");
			Y -= dy;
			add_sound(SND_WALL1);
			add_sound(SND_WALL2);
			damage(Task, Speed); }
		c = peek(arena_seg, (X*Y_ARENA)|Y);
		if(c & 0x80) {
			message("Collided with %u", (c & 0x7F)+1);
			X -= dx;
			Y -= dy;
			add_sound(SND_CRASH1);
			add_sound(SND_CRASH2);
			damage(Task, Speed);
			damage(c & 0x7F, Speed + 5); } }
}

/*
 * Test for a robot within an area location
 */
is_here(int x, int y, int r, int a)
{
	return (abs(CF_block[r]._X - x) <= a) && (abs(CF_block[r]._Y - y) <= a);
}

/* Add a shot */
add_shot(unsigned direction)
{
	int x;

	if(!Shot) {			/* Is robot able to shoot now? */
		add_sound(SND_SHOT);
		for(x = 0; x < MAX_SHOTS; ++x)		/* find spot to add shot */
			if(!shots_v[x]) {
				Shot = Shot_delay;
				shots_x[x] = X;
				shots_y[x] = Y;
				shots_d[x] = direction;
				shots_v[x] = Shot_dist;
				shots_o[x] = Task;
				return; } }
}

/* Move shots and do damage */
do_shots()
{
	int x, y, s;
	char hit;

	/* move shots and do damage */
	for(x = 0; x < MAX_SHOTS; ++x)
		if(shots_v[x]) {
			w_gotoxy(shots_x[x] / (X_ARENA/X_SIZE), shots_y[x] / (Y_ARENA/Y_SIZE), Awin);
			w_putc(' ', Awin);

			shots_x[x] += delta_x[shots_d[x]] * 2;
			shots_y[x] += delta_y[shots_d[x]] * 2;

			/* out of bounds? */
			if((shots_x[x] >= X_ARENA) || (shots_y[x] >= Y_ARENA))
					shots_v[x] = 0;
			else {
				hit = 0;
				for(y = 0; y < CF_tasks; ++y)
					if((shots_o[x] != y) && is_here(shots_x[x], shots_y[x], y, Shot_area)) {
						wprintf("\n[%u] Scored hit on %u", s = shots_o[x]+1, y+1);
						if(mfp)
							fprintf(mfp, "[%u] Scored hit on %u\n", s, y+1);
						add_sound(SND_HIT);
						damage(y, shots_v[x]);
						hit = -1; }
				shots_v[x] = hit ? 0 : shots_v[x]-1; } }

	/* draw all shots */
	for(x = 0; x < MAX_SHOTS; ++x)
		if(shots_v[x]) {
			w_gotoxy(shots_x[x] / (X_ARENA/X_SIZE), shots_y[x] / (Y_ARENA/Y_SIZE), Awin);
			w_putc('*', Awin); }
}

/*
 * Scan in a direction for a robot... If found, return the ID number
 * (1+) of the robot in question.
 */
scan(unsigned dir)
{
	unsigned x, y, i, j;

	/* get starting point */
	x = X; y = Y;

	for(i = 0; i < Scan_dist; ++i) {
		/* move along direction */
		x += delta_x[dir];
		y += delta_y[dir];

		if((x >= X_ARENA) || (y >= X_ARENA))
			return 0;

		/* scan */
		for(j = 0; j < CF_tasks; ++j) {
			if((j != Task) && is_here(x, y, j, Scan_area)) {
				if(!CF_block[j]._PC)
					continue;
				return j+1; } } }

	return 0;
}

/*
 * Add a sound to the sound queue
 */
add_sound(unsigned sound)
{
	if(Sound) {
		Sound_queue[Sound_head] = sound;
		Sound_head = (Sound_head + 1) & (SND_QUEUE-1); }
}

/* --- Main program --- */

/*
 * Parse arguments, load files, and control robots in arena.
 */
main(int argc, char *argv[])
{
	int i;
	unsigned lp, t, wt, wc;
	static int starting_positions[MAX_TASKS][2] = {
		X_START,			Y_START,			/* Upper left corner */
		X_ARENA-X_START,	Y_ARENA-Y_START,	/* Lower right corner */
		X_START,			Y_ARENA-Y_START,	/* Lower Left corner */
		X_ARENA-X_START,	Y_START,			/* Upper right corner */
		X_ARENA/2,			Y_START,			/* Upper center */
		X_ARENA/2,			Y_ARENA-Y_START,	/* Lower Center */
		X_START*2,			Y_ARENA/2,			/* Left center */
		X_ARENA-(X_START*2),Y_ARENA/2 };		/* Right center */

	if(argc < 2)
		abort(help);

	if(!(arena_seg = alloc_seg((X_ARENA/16)*Y_ARENA)))
		abort("Cannot allocate memory");

	RAND_SEED = peekw(0x40, 0x6C);

	lp = X_ARENA * Y_ARENA;
	do {
		pokew(arena_seg, lp -= 2, 0); }
	while(lp);

	wopen(0, 0, 1, 1, 0);
	wclose();
	if(W_BASE == 0xB000)
		attrs = mattrs;

	Awin = wopen(0, 0, X_SIZE, Y_SIZE, attrs[0]);
	W_OPEN->WINpcurx = W_OPEN->WINpcury = 0;
	wcursor_off();
	Swin = wopen(X_SIZE, 0, 80-X_SIZE, Y_SIZE, attrs[1]);
	wprintf(" *** DAMAGE ***");
	Mwin = wopen(0, Y_SIZE, 80, 25-Y_SIZE, attrs[2]);
	wprintf("Micro-C ROBOTS - Copyright 1999-2005 Dave Dunfield.");

	memset(shots_x, 0, sizeof(shots_x));
	memset(CF_block, 0, sizeof(CF_block));

	for(i=1; i < argc; ++i) {
		ptr = argv[i];
		switch((toupper(*ptr++) << 8) | toupper(*ptr++)) {
		case '-I' : Inter = 0;					continue;
		case '-R' : Rstart = 0;					continue;
		case '-S' : Sound = 0;					continue;
		case '-W' : Win = 0; 					continue;
		case 'A=' : Scan_area = atoi(ptr);		continue;
		case 'C=' : Scan_dist = atoi(ptr);		continue;
		case 'F=' : mfp = fopen(ptr, "wvq");	continue;
		case 'H=' : Shot_dist = atoi(ptr);		continue;
		case 'I=' : Slottime = atoi(ptr);		continue;
		case 'M=' : Maxdamage = atoi(ptr);		continue;
		case 'O=' : Shot_area = atoi(ptr);		continue;
		case 'T=' : Shot_delay = atoi(ptr); 	continue; }

		if(CF_tasks >= MAX_TASKS)
			error("Too many robots");

		CF_ptr = CF_block[Task = CF_tasks];
		if(!(CFSEG = alloc_seg(4096)))
			error("Cannot allocate memory");
		lp = read_file(argv[i]);
		Name = argv[i];
		message("%u bytes loaded at segment %04x, PC=%04x",
			lp-PC, CFSEG, PC);
		w_gotoxy(1, ++CF_tasks, Swin);
		w_printf(Swin,"%u:%-8s:", CF_tasks, Name); }

	switch(CF_tasks) {
		case 0:	error("You must specify at least one robot file.");
		case 1: Win = 0; }

	for(Task = 0; Task < CF_tasks; ++Task) {
		CF_ptr = CF_block[Task];
		wt = starting_positions[Task][0];
		wc = starting_positions[Task][1];
		if(Rstart) {
			wt = (wt - X_START) + random(X_START*2);
			wc = (wc - Y_START) + random(Y_START*2); }
		X = wt;
		Y = wc; }

	if(Inter) {
		wform(20, 5, attrs[3], data_form,
			&Slottime, &Maxdamage, &Scan_dist, &Scan_area, &Shot_dist,
			&Shot_area, &Shot_delay); }

	Task = t = 0;
	wc = 1000;
	for(;;) {
		if(++Task >= CF_tasks) {
			sound_off();
			if(kbtst() == 0x1B) {
				wc = -1;
				break; }
			if((wc < 2) && (Sound_head == Sound_tail) && Win)
				break;
			if(Sound_head != Sound_tail) {
				sound(Sound_queue[Sound_tail]);
				Sound_tail = (Sound_tail + 1) & (SND_QUEUE-1); }
			Task = wc = 0;
			do_shots();
			delay(100);
			if(++Tick > 9)
				Tick = 0;
			if(++t >= CF_tasks)
				t = 0;
			w_gotoxy(12, t+1, Swin);
			w_printf(Swin,"%4u", CF_block[t]._Damage); }
		CF_ptr = CF_block[Task];
		if(PC) {
			wt = Task;
			++wc;
			cf_run();
			move_robot();
			w_gotoxy(X / (X_ARENA/X_SIZE), Y / (Y_ARENA/Y_SIZE), Awin);
			w_putc(Task+'1', Awin);
			poke(arena_seg, (X*Y_ARENA)|Y, Task|0x80);
			if(Shot)
				--Shot; }
		else {
			w_gotoxy(X / (X_ARENA/X_SIZE), Y / (Y_ARENA/Y_SIZE), Awin);
			*Awin = REVERSE;
			w_putc(Task+'1', Awin);
			*Awin = *attrs;
			poke(arena_seg, (X*Y_ARENA)|Y, Task|0x80); } }

	if(Inter) {
		w_clwin(Awin);
		for(Task=0; Task < CF_tasks; ++Task) {
			CF_ptr = CF_block[Task];
			w_printf(Awin, "%u:%-8s: PC: %04x Damage:%-4u Direction:%-2u Speed:%u\n",
				Task+1, Name, PC, Damage, Direct, Speed); }

		if(Win) switch(wc) {
			case 0 :
				w_printf(Awin, "\nAll robots died!\n");
				break;
			case 1 :
				w_printf(Awin, "\nAnd the winner is: %u\n", wt+1);
				break;
			default:
				w_printf(Awin, "\nRobots were still fighting!\n"); }
		w_puts("\nPress any key to exit.", Awin);
		wgetc(); }

	*Mwin = *Swin = *Awin = NORMAL;
	wclose();
	wclose();
	wclose();

	if(!Inter) {
		for(Task=0; Task < CF_tasks; ++Task) {
			CF_ptr = CF_block[Task];
			printf("%u:%-8s: PC: %04x Damage:%-4u Direction:%-2u Speed:%u\n",
				Task+1, Name, PC, Damage, Direct, Speed); }

		if(Win) switch(wc) {
			case 0 :
				printf("\nAll robots died!\n");
				break;
			case 1 :
				printf("\nAnd the winner is: %u\n", wt+1);
				break;
			default:
				printf("\nRobots were still fighting!\n"); } }

	if(mfp) {
		for(Task=0; Task < CF_tasks; ++Task) {
			CF_ptr = CF_block[Task];
			fprintf(mfp, "%u:%-8s: PC: %04x Damage:%-4u Direction:%-2u Speed:%u\n",
				Task+1, Name, PC, Damage, Direct, Speed); }

		if(Win) switch(wc) {
			case 0 :
				fprintf(mfp, "\nAll robots died!\n");
				break;
			case 1 :
				fprintf(mfp, "\nAnd the winner is: %u\n", wt+1);
				break;
			default:
				fprintf(mfp, "\nRobots were still fighting!\n"); }
		fclose(mfp); }
}
