/*
 * Example MICRO-C/PC program showing how to use the VMPC86 module
 * to run a C-FLEA program within an application program written
 * in MICRO-C for the PC.
 *
 * To compile this program:
 *		masm/ml VMPC86;			<- Assemble VM module
 *		cc RUNCF -flmop			<- Compile this to object
 *		lc RUNCF VMPC86			<- Link together with VMPC86
 *
 * To compile the C-FLEA program:
 *		CCF program -<options>	<- Compile program as usual
 *		HEXFMT program.hex -bt	<- Convert to binary
 *
 * To load and execute the CFLEA program:
 *		CFRUN program.hex
 *
 * Copyright 1996-2005 Dave Dunfield
 * All rights reserved.
 */
#include <stdio.h>
// Make sure LOADADR matches the starting address configured in library
#define	LOADADR	0x1000			// Default load address is 0x1000

// C-FLEA variables defined in VMPC86
extern unsigned CFSEG,			// Segment where C-FLEA memory is located
	ACC,						// C-FLEA Accumulator
	INDEX,						// C-FLEA Index register
	PC,							// C-FLEA Program Counter
	SP;							// C-FLEA Stack pointer

// C-FLEA functions defined in VMPC86
extern int cf_run(void);		// Executes CFLEA program

// C-FLEA functions required by VMPC86 (supplied in this module)
void cf_output(unsigned port);	// Called to output to an I/O port
void cf_input(unsigned port);	// Called to input from an I/O port

/*
 * Main program
 */
main(int argc, char *argv[])
{
	int i;
	FILE *fp;
	static unsigned lp = LOADADR;	// Load program here

	if(argc < 2)
		abort("\nUse: runcf <file>\n");

	fp = fopen(argv[1], "rvqb");	// Open binary image file

	if(!(CFSEG = alloc_seg(4096)))	// Allocate 64K run segment
		abort("Cannot allocate memory");

	while((i = getc(fp)) >= 0)		// Load in executable
		poke(CFSEG, lp++, i);

	fclose(fp);						// Finished with file

	printf("C-FLEA: %u bytes loaded at segment %04x\n", lp - LOADADR, CFSEG);

	PC = LOADADR;					// Set initial execution address

	// Run C-FLEA program, showing how to implement our own instructions.
	// The secret is that cf_run() stops on any unrecognized opcode, and
	// returns the opcode value to us. Here we implement an instruction
	// which dumps all C-FLEA registers to the console (for debugging).
runcf:
	switch(i = cf_run()) {
		case 0xED :			// Special debug instruction
			printf("C-FLEA DEBUG: ACC=%04x INDEX=%04x PC=%04x SP=%04x\n",
				ACC, INDEX, PC, SP);
			++PC;			// Skip this opcode
			goto runcf;		// And keep going
		// More instructions can be defined here
		}

	printf("C-FLEA: Exit at %04x, opcode: %02x\n", PC, i);

	free_seg(CFSEG);
}

/*
 * This function is called in response to an OUT instruction
 */
void cf_output(unsigned port)
{
	switch(port) {
		case 0 :	// Console output
			putc(ACC, stdout);
			return;
		// More output ports can be defined here
		}

	printf("\nC-FLEA: Unknown output port %02x at %04x\n", port, PC-2);
	exit(-1);
}

/*
 * This function is called in response to an IN instruction
 */
void cf_input(unsigned port)
{
	switch(port) {
		case 0 :	// Console input
			ACC = kbget();
			return;
		// More input ports can be defined here
		}

	printf("\nC-FLEA: Unknown input port %02x at %04x\n", port, PC-2);
	exit(-1);
}
