/*
 * C-FLEA Virtual Machine implementation in C.
 *
 * This provides a portable C-FLEA implementation which can be
 * included in any 'C' program.
 *
 * This program can be compiled with DDS Micro-C/PC, or with Borland's
 * TURBO-C, which is fairly representative of a generic 16 bit DOS
 * compiler. Look for the "#ifdef's" to note portability issues.
 *
 * To compile this program with Micro-C/PC:
 *		CC VMC -pof				<- Compile to VMC.COM
 *
 * To compile this program with TURBO-C 2.0:
 *		TCC -mt VMC
 *		exe2bin VMC.EXE VMC.COM
 *
 * To compile the C-FLEA program:
 *		CCF program -<options>	<- Compile program as usual
 *		HEXFMT program.hex -bt	<- Convert to binary
 *
 * To load and execute the CFLEA program:
 *		VMC program.hex
 *
 * Copyright 1996-2005 Dave Dunfield
 * All rights reserved.
 */
#include <stdio.h>
/* Make sure LOADADR matches the starting address configured in library */
#define	LOADADR	0x1000		/* Default load address is 0x1000 */

#define	BYTE_OP	0x08		/* Bit in opcode indicating byte operand */
#define	EQ		0x01		/* Bit in cflags indicating == */
#define	ULT		0x02		/* Bit in cflags indicating unsigned < */
#define	UGT		0x04		/* Bit in cflags indicating unsigned > */
#define	SLT		0x08		/* Bit in cflags indicating signed < */
#define	SGT		0x10		/* Bit in cflags indicating signed > */

/*
 * Public global variables
 */
unsigned
	ACC,					/* C-FLEA accumulator */
	INDEX,					/* C-FLEA Index register */
	SP,						/* C-FLEA Stack Pointer */
	PC,						/* C-FLEA Program Counter */
	CFSEG;					/* C-FLEA code image segment (if used) */

/*
 * Private global variables
 */
static unsigned
	Alt,					/* Alternate result from DIV */
	Address;				/* Address holding register */
static unsigned char
	opcode,					/* current command opcode */
	cflags;					/* Compare result flags */

/*
 * Function prototypes
 */
unsigned cf_run(void);			/* Main C-FLEA execution function */
void cf_output(unsigned port);	/* Called to output to an I/O port */
void cf_input(unsigned port);	/* Called to input from an I/O port */

/*
 * Functions to read and write C-FLEA memory
 *
 * For this demo, we will simply allocate a 64k segment and use the
 * host compilers peek/poke functions to access it, however you may
 * implement any memory and addressing scheme you wish.
 */
#ifdef _MICROC_
	#define	read_mem(address)			peek(CFSEG, address)
	#define	write_mem(address, data)	poke(CFSEG, address, data)
#else
	unsigned char peekb(unsigned segment, unsigned offset);
	void pokeb(unsigned segment, unsigned offset, int value);
	#define	read_mem(address)			peekb(CFSEG, address)
	#define	write_mem(address, data)	pokeb(CFSEG, address, data)
#endif

/*
 * Compute the address of the current instructions operand
 */
static void cf_get_address(void)
{
	switch(opcode & 7) {
		case 0 :		/* Immediate reference */
			Address = PC++;
			return;
		case 1 :		/* Direct address */
			Address = read_mem(PC++);
			Address |= read_mem(PC++) << 8;
			return;
		case 2 :		/* Indirect - no offset */
			Address = INDEX;
			return;
		case 3 :		/* Indirect with offset */
			Address = INDEX + read_mem(PC++);
			return;
		case 4 :		/* Indirect from SP with offset */
			Address = SP + read_mem(PC++);
			return;
		case 5 :		/* On top of stack, remove */
			Address = SP;
			SP += 2;
			return;
		case 6 :		/* Indirect through top of stack, remove */
			Address = read_mem(SP++);
			Address |= read_mem(SP++) << 8;
			return;
		case 7 :
			Address = read_mem(SP);
			Address |= read_mem(SP+1) << 8;
			return; }
}

/*
 * Main C-FLEA virtual machine execution function
 */
unsigned cf_run(void)
{
	unsigned operand, t1, t2;

	for(;;) {
		opcode = read_mem(PC++);

		if(opcode < 0x98) {		/* Catagory 1 instruction */
			cf_get_address();
			operand = read_mem(Address);
			if(!(opcode & BYTE_OP)) {
				if(!(opcode & 7))
					++PC;
				operand |= read_mem(Address+1) << 8; }
			switch(opcode & 0xF0) {
			case 0x00 :	ACC = operand;					break;	/* LD */
			case 0x10 : ACC += operand;					break;	/* ADD */
			case 0x20 : ACC -= operand;					break;	/* SUB */
			case 0x30 : ACC *= operand;					break;	/* MUL */
			case 0x40 :
				Alt = ACC % operand; ACC /= operand;	break;	/* DIV */
			case 0x50 : ACC &= operand;					break;	/* AND */
			case 0x60 :	ACC |= operand;					break;	/* OR */
			case 0x70 :	ACC ^= operand;					break;	/* XOR */
			case 0x80 :											/* CMP */
				if(ACC == operand) {
					cflags = EQ;
					ACC = 1; }
				else {
					cflags = (ACC < operand) ? ULT : UGT;
					cflags |= ((int)ACC < (int)operand) ? SLT : SGT;
					ACC = 0; }
				break;
			case 0x90 :	INDEX = operand; }						/* LDI */
		} else if (opcode < 0xC8) {		/* Catagory 2 instruction */
			cf_get_address();
			switch(opcode & 0xF8) {
			case 0x98 :	INDEX = Address;				break;	/* LEAI */
			case 0xA0 :	write_mem(Address+1, ACC >> 8);			/* ST */
			case 0xA8 :	write_mem(Address, ACC);		break;	/* STB */
			case 0xB0 :	write_mem(Address, INDEX);				/* STI */
						write_mem(Address+1, INDEX >> 8); break;
			case 0xB8 : ACC >>= read_mem(Address);		break;	/* SHR */
			case 0xC0 :	ACC <<= read_mem(Address); }			/* SHL */
		} else switch(opcode) {		/* Catagory 3 instruction */
		case 0xC8 :	ACC = (cflags & (SLT))	  ? 1 : 0;	break;	/* LT */
		case 0xC9 : ACC = (cflags & (SLT|EQ)) ? 1 : 0;	break;	/* LE */
		case 0xCA : ACC = (cflags & (SGT))	  ? 1 : 0;	break;	/* GT */
		case 0xCB : ACC = (cflags & (SGT|EQ)) ? 1 : 0;	break;	/* GE */
		case 0xCC : ACC = (cflags & (ULT))	  ? 1 : 0;	break;	/* ULT */
		case 0xCD : ACC = (cflags & (ULT|EQ)) ? 1 : 0;	break;	/* ULE */
		case 0xCE : ACC = (cflags & (UGT))	  ? 1 : 0;	break;	/* UGT */
		case 0xCF : ACC = (cflags & (UGT|EQ)) ? 1 : 0;	break;	/* UGE */
		case 0xD0 :												/* JMP */
		doljmp:		Address = read_mem(PC++);
					PC = (read_mem(PC++) << 8) | Address;break;
		case 0xD1 : if(!ACC) goto doljmp; PC+=2;		break;	/* JZ */
		case 0xD2 :	if(ACC)  goto doljmp; PC+=2;		break;	/* JNZ */
		case 0xD3 :
		dosjmp:		Address = read_mem(PC++);					/* SJMP */
					if(Address & 0x80) Address |= 0xFF00;
					PC += Address;						break;
		case 0xD4 :	if(!ACC) goto dosjmp; ++PC;			break;	/* SJZ */
		case 0xD5 :	if(ACC)  goto dosjmp; ++PC;			break;	/* SJNZ */
		case 0xD6 :	PC = ACC;							break;	/* IJMP */
		case 0xD7 :												/* SWITCH */
			Address = INDEX;
			for(;;) {
				t1 = read_mem(Address++);
				t1 |= read_mem(Address++) << 8;
				t2 = read_mem(Address++);
				t2 |= read_mem(Address++) << 8;
				if(!t1) {
					PC = t2;
					break; }
				if(t2 == ACC) {
					PC = t1;
					break; } }
			break;
		case 0xD8 :												/* CALL */
			operand = PC + 2;
			write_mem(--SP, operand >> 8);
			write_mem(--SP, operand);
			goto doljmp;
		case 0xD9 :												/* RET */
			PC = read_mem(SP++);
			PC |= read_mem(SP++) << 8;					break;
		case 0xDA : SP -= read_mem(PC++);				break;	/* ALLOC */
		case 0xDB :	SP += read_mem(PC++);				break;	/* FREE */
		case 0xDC :												/* PUSHA */
			write_mem(--SP, ACC >> 8);
			write_mem(--SP, ACC);						break;
		case 0xDD :												/* PUSHI */
			write_mem(--SP, INDEX >> 8);
			write_mem(--SP, INDEX);						break;
		case 0xDE : SP = ACC;							break;	/* TAS */
		case 0xDF : ACC = SP;							break;	/* TSA */
		case 0xE0 : ACC = 0;							break;	/* CLR */
		case 0xE1 : ACC = ~ACC;							break;	/* COM */
		case 0xE2 : ACC = -ACC;							break;	/* NEG */
		case 0xE3 : ACC = !ACC;							break;	/* NOT */
		case 0xE4 : ++ACC;								break;	/* INC */
		case 0xE5 : --ACC;								break;	/* DEC */
		case 0xE6 : INDEX = ACC;						break;	/* TAI */
		case 0xE7 : ACC = INDEX;						break;	/* TIA */
		case 0xE8 : INDEX += ACC;						break;	/* ADAI */
		case 0xE9 : ACC = Alt;							break;	/* ALT */
		case 0xEA : cf_output(read_mem(PC++));			break;	/* OUT */
		case 0xEB : cf_input(read_mem(PC++));			break;	/* IN */
		default: return opcode; } }			/* Unknown opcode */
}

/*
 * This function is called in response to an OUT instruction
 */
void cf_output(unsigned port)
{
	switch(port) {
		case 0 :	/* Console output */
			putc(ACC, stdout);
			return;
		/* More output ports can be defined here */
		}

	printf("\nC-FLEA: Unknown output port %02x at %04x\n", port, PC-2);
	exit(-1);
}

/*
 * This function is called in response to an IN instruction
 */
void cf_input(unsigned port)
{
	switch(port) {
		case 0 :	/* Console input */
#ifdef _MICROC_
			ACC = kbget();
#else
			ACC = getch();
#endif
			return;
		/* More input ports can be defined here */
		}

	printf("\nC-FLEA: Unknown input port %02x at %04x\n", port, PC-2);
	exit(-1);
}

/*
 * Demo program to show use of the VMC module.
 */
main(int argc, char *argv[])
{
	int i;
	FILE *fp;
	static unsigned lp = LOADADR;	/* Load program here */

	if(argc < 2)
		abort("\nUse: vmc <file>\n");

	if(!(fp = fopen(argv[1], "rb"))) {	/* Open binary image file */
		printf("Cannot open input file");
		return; }

#ifdef _MICROC_
	if(!(CFSEG = alloc_seg(4096)))		/* Allocate 64K run segment */
		abort("Cannot allocate memory");
#else
	if(allocmem(4096, &CFSEG) != -1) {	/* Allocate 64K run segment */
		printf("Cannot allocate memory");
		return; }
#endif

	while((i = getc(fp)) >= 0)		/* Load in executable image */
		write_mem(lp++, i);

	fclose(fp);						/* Finished with file */

	printf("C-FLEA: %u bytes loaded at segment %04x\n", lp - LOADADR, CFSEG);

	PC = LOADADR;					/* Set initial execution address */

	/* Run C-FLEA program, showing how to implement our own instructions.
	 * The secret is that cf_run() stops on any unrecognized opcode, and
	 * returns the opcode value to us. Here we implement an instruction
	 * which dumps all C-FLEA registers to the console (for debugging). */
runcf:
	switch(i = cf_run()) {
		case 0xED :			/* Special debug instruction */
			printf("C-FLEA DEBUG: ACC=%04x INDEX=%04x PC=%04x SP=%04x\n",
				ACC, INDEX, PC, SP);
			goto runcf;		/* And keep going */
		/* More instructions can be defined here */
		}

	printf("C-FLEA: Exit at %04x, opcode: %02x\n", PC-1, i);

#ifdef _MICROC_
	free_seg(CFSEG);
#else
	freemem(CFSEG);
#endif
}
